"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const oracledb = require("oracledb");
const base_1 = require("../error/base");
const db_service_1 = require("./db_service");
const db_1 = require("../error/db");
exports.DB_CONNECT_ERROR = 'DB_QUERY';
exports.DB_ORACLE_ERROR = 'DB_ORACLE_ERROR';
exports.DB_ORACLE_FETCH_ERROR = 'DB_ORACLE_FETCH_ERROR';
exports.DB_ORACLE_CLOSE_ERROR = 'DB_ORACLE_CLOSE_ERROR';
exports.DB_ORACLE_RELEASE_ERROR = 'DB_ORACLE_RELEASE_ERROR';
exports.DB_ORACLE_EXECUTE_ERROR = 'DB_ORACLE_EXECUTE_ERROR';
class OracleService {
    constructor(connectionParams, options) {
        this.poolMax = 100;
        this.maxQueryAttempts = 10;
        this._queryFailsCounter = {};
        this.options = options || {};
        this.connectionParams = connectionParams;
    }
    run(query, params, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this._run(query, params, options);
        });
    }
    getManyRows(query, params = []) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                return yield this._run(query, params, { resultSet: true, prefetchRows: 500 });
            }
            catch (error) {
                throw new db_1.DbError({
                    code: exports.DB_ORACLE_ERROR,
                    innerDetails: {
                        query,
                        message: error.message,
                        params
                    }
                });
            }
        });
    }
    fetchRows(resultSet, numRows) {
        return __awaiter(this, void 0, void 0, function* () {
            let rows;
            try {
                rows = yield resultSet.getRows(numRows);
            }
            catch (error) {
                yield this.closeResultSet(resultSet);
                throw new db_1.DbError({
                    code: exports.DB_ORACLE_FETCH_ERROR,
                    innerDetails: error.message
                });
            }
            if (rows.length === 0) {
                yield this.closeResultSet(resultSet);
                return [];
            }
            return rows;
        });
    }
    closeResultSet(resultSet, connection) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield resultSet.close();
            }
            catch (error) {
                throw new db_1.DbError({
                    code: exports.DB_ORACLE_FETCH_ERROR,
                    innerDetails: error.message
                });
            }
            if (connection) {
                yield this._disconnect(connection);
            }
        });
    }
    getRows(query, params = [], options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const result = yield this._run(query, params, options);
                return result.rows;
            }
            catch (error) {
                throw new db_1.DbError({
                    code: exports.DB_ORACLE_ERROR,
                    innerDetails: {
                        query,
                        message: error.message,
                        params
                    }
                });
            }
        });
    }
    getRow(query, params = [], options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const rows = yield this.getRows(query, params, options);
            if (rows.length === 0) {
                return false;
            }
            if (rows.length > 1) {
                console.warn(db_service_1.ONE_ROW_WARNING, rows.length, query);
            }
            return rows[0];
        });
    }
    mustGetRow(errorCode, query, params = [], options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const row = yield this.getRow(query, params, options);
            if (row === false) {
                throw new db_1.DbError({
                    code: errorCode,
                    status: base_1.BaseError.CODE_NOT_FOUND,
                    innerDetails: {}
                });
            }
            return row;
        });
    }
    _connect() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                if (!this.pool) {
                    this.pool = yield oracledb.createPool(Object.assign({}, this.connectionParams, { poolMax: this.poolMax }));
                }
                return yield this.pool.getConnection();
            }
            catch (error) {
                throw new db_1.DbError({
                    code: exports.DB_CONNECT_ERROR,
                    innerDetails: error.message
                });
            }
        });
    }
    _disconnect(connection) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield connection.close();
            }
            catch (error) {
                throw new db_1.DbError({
                    code: exports.DB_ORACLE_RELEASE_ERROR,
                    innerDetails: error.message
                });
            }
        });
    }
    _run(query, params, options) {
        return __awaiter(this, arguments, void 0, function* () {
            let connection;
            options = Object.assign({}, this.options, options);
            try {
                connection = yield this._connect();
                const result = yield connection.execute(query, params, options);
                yield this._disconnect(connection);
                return result;
            }
            catch (error) {
                if (connection) {
                    yield this._disconnect(connection);
                }
                if (this._needRetry(arguments, error)) {
                    yield this.pool.close();
                    this.pool = null;
                    return this._run(query, params, options);
                }
                else {
                    throw new db_1.DbError({
                        publicMessage: error.message,
                        innerDetails: error
                    });
                }
            }
        });
    }
    _needRetry(runArguments, error) {
        if (!error || !error.message) {
            return false;
        }
        const retryErrorsRegexp = /^(ORA-03113|ORA-04068|ORA-00028|ORA-02020)/ig;
        if (!retryErrorsRegexp.test(error.message)) {
            return false;
        }
        const failsCounterKey = JSON.stringify(runArguments);
        if (!this._queryFailsCounter[failsCounterKey]) {
            this._queryFailsCounter[failsCounterKey] = 0;
        }
        this._queryFailsCounter[failsCounterKey]++;
        return this._queryFailsCounter[failsCounterKey] < this.maxQueryAttempts;
    }
}
exports.OracleService = OracleService;
